#ifdef HAVE_CONFIG_H
#include <config.h>
#endif /* HAVE_CONFIG_H */

#include <tarot.h>
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>

#ifndef LOCALEDIR
#define LOCALEDIR "/usr/share/locale"
#endif /* NOT LOCALEDIR */

int
main ()
{
  TarotGame *game;
  TarotGameEvent *event;
  const TarotGameEvent *read_only_event;
  static const char *deal_card_names[] =
    {
     "10H", "JH", "CH", "QH", "KH",
     "11T", "12T", "13T", "14T", "15T", "16T", "17T", "18T", "19T",
     "EXC"
    };
  TarotCard deal_cards[sizeof (deal_card_names) / sizeof (deal_card_names[0])];
  size_t i;
  TarotGameIterator *iterator;
  TarotCard *dog_cards;
  size_t n_cards;
  
  if (tarot_init (LOCALEDIR) != 0)
    {
      fprintf (stderr, "Error: could not initialize libtarot.\n");
      return EXIT_FAILURE;
    }
  game = tarot_game_alloc ();
  
  /* The game has 5 players, with call */
  event = tarot_game_event_alloc_setup (5, 1);
  if (tarot_game_add_event (game, event) != TAROT_GAME_OK)
    {
      fprintf (stderr, "The game refused to be set up for 5 players, with call!\n");
      return EXIT_FAILURE;
    }
  tarot_game_event_free (event);

  /* Deal the cards */
  for (i = 0; i < sizeof (deal_cards) / sizeof (deal_cards[0]); i++)
    {
      deal_cards[i] = tarot_string_to_card_c (deal_card_names[i]);
      assert (deal_cards[i] != ((TarotCard) (-1)));
    }
  event = tarot_game_event_alloc_deal (tarot_string_to_player_c ("P5"),
				       sizeof (deal_cards) / sizeof (deal_cards[0]),
				       deal_cards);
  if (tarot_game_add_event (game, event) != TAROT_GAME_OK)
    {
      fprintf (stderr, "The game refused to deal cards to P5!\n");
      return EXIT_FAILURE;
    }
  tarot_game_event_free (event);

  /* Auto-reveal */
  if (tarot_game_can_autoreveal_alloc (game, &n_cards, &dog_cards))
    {
      /* This is a game for which we know the dog and we should add it! */
      event = tarot_game_event_alloc_dog (n_cards, dog_cards);
      free (dog_cards);
      if (tarot_game_check_event (game, event))
	{
	  printf ("The game is effectively waiting for the dog.\n");
	}
      if (tarot_game_add_event (game, event) != TAROT_GAME_OK)
	{
	  fprintf (stderr, "Could not auto-reveal the dog!\n");
	  return EXIT_FAILURE;
	}
      tarot_game_event_free (event);
    }

  /* Iterate over the events */
  iterator = tarot_game_iterator (game);
  while ((read_only_event = tarot_game_iterator_next_value (iterator)) != NULL)
    {
      switch (tarot_game_event_type (read_only_event))
	{
	case TAROT_SETUP_EVENT:
	  {
	    size_t n_players;
	    int with_call;
	    if (tarot_game_event_get_setup (read_only_event, &n_players, &with_call) != TAROT_EVENT_OK)
	      {
		fprintf (stderr, "This is not possible.\n");
		return EXIT_FAILURE;
	      }
	    printf ("There is a setup event!  The game has %lu players (with call? %d)\n",
		    n_players, with_call);
	  }
	  break;
	case TAROT_DEAL_EVENT:
	  {
	    TarotCard *cards;
	    TarotPlayer who;
	    char *player_name, *card_1_name, *card_2_name;
	    TarotCard card2;
	    size_t has_card_2;
	    if (tarot_game_event_get_deal_alloc (read_only_event, &who, &n_cards, &cards) != TAROT_EVENT_OK)
	      {
		return EXIT_FAILURE;
	      }
	    if (n_cards < 2)
	      {
		return EXIT_FAILURE;
	      }
	    /* Query only a subset (of size 1... i.e. only one card)
	       from the deal event */
	    if (tarot_game_event_get_deal (read_only_event, &who, &has_card_2, 1, 1, &card2) != TAROT_EVENT_OK)
	      {
		return EXIT_FAILURE;
	      }
	    assert (has_card_2 > 0);
	    assert (card2 == cards[1]);
	    player_name = tarot_player_to_string_alloc (who);
	    card_1_name = tarot_card_to_string_alloc (cards[0]);
	    card_2_name = tarot_card_to_string_alloc (cards[1]);
	    printf ("There is a deal event!  The lucky player %s receives cards %s, %s, ...\n",
		    player_name, card_1_name, card_2_name);
	    free (player_name);
	    free (card_1_name);
	    free (card_2_name);
	    free (cards);
	  }
	  break;
	default:
	  fprintf (stderr, "Unknown game event.\n");
	  return EXIT_FAILURE;
	}
    }
  tarot_game_iterator_free (iterator);
  tarot_game_free (game);
  tarot_quit ();
  return EXIT_SUCCESS;
}
